import { Request, Response } from "express";
import imageService from "../services/ImageService";
import mongoose, { mongo } from "mongoose";
import { Image } from "../models/ImageModel";

class ImageController {
  async editImage(req: Request, res: Response) {
    try {
      const change: Image = req.body;
      const result = await imageService.replaceOne(req.params.id, change);
      if (result.matchedCount > 0) {
        res.status(204).json();
      } else {
        res.status(404).json({ error: "Image not found" });
      }
    } catch (error: any) {
      if (error instanceof mongoose.Error.CastError) {
        res.status(400).json({ error: "Invalid Id" });
      } else if (error instanceof mongoose.Error.ValidationError) {
        // Should return 400 Bad request for invalid requests
        res.status(400).json({ error: error.message });
      } else if (
        error instanceof mongo.MongoServerError &&
        error.code === 11000
      ) {
        // Should return 409 Conflict for existing urls
        res.status(409).json({
          error: `the image with URL ${error.keyValue.url} already exists`,
        });
      } else {
        res.status(500).json({ error: "Internal Server Error" });
      }
    }
  }
  async getImageById(req: Request, res: Response) {
    try {
      const image = await imageService.findById(req.params.id);
      if (image) {
        res.json({ image });
      } else {
        res.status(404).json({ error: "Image not found" });
      }
    } catch (error: any) {
      if (error instanceof mongoose.Error.CastError) {
        res.status(400).json({ error: "Invalid Id" });
      } else {
        res.status(500).json({ error: "Internal Server Error" });
      }
    }
  }

  async getAllImages(req: Request, res: Response): Promise<void> {
    try {
      if (
        req.query.status !== undefined &&
        !["consumed", "unavailable", "available"].includes(
          req.query.status as string
        )
      ) {
        throw TypeError(
          "if present, `status` should have the values `consumed`, `unavailable`, or `available`"
        );
      }
      const limit = req.query.limit ? Number(req.query.limit) : undefined;
      const status = req.query.status
        ? (req.query.status as Image["status"])
        : undefined;

      const images = await imageService.findAll(limit, status);
      res.json({ images });
    } catch (error) {
      if (error instanceof TypeError) {
        res.status(400).json({ error });
      } else {
        res.status(500).json({ error: "Internal Server Error" });
      }
    }
  }

  async addImage(req: Request, res: Response): Promise<void> {
    try {
      // Should add auth here before doing stuff
      // Thowing a 401 if not auth provided
      // Throwing a 403 for incorrect auth
      const image = await imageService.add(req.body);
      res.status(201).json({ image });
    } catch (error: any) {
      if (error instanceof mongo.MongoServerError && error.code === 11000) {
        // Should return 409 Conflict for existing urls
        res.status(409).json({
          error: `the image with URL ${error.keyValue.url} already exists`,
        });
      } else if (error instanceof mongoose.Error.ValidationError) {
        // Should return 400 Bad request for invalid requests
        res.status(400).json({ error: error.message });
      } else {
        // Return 500 in other case
        res.status(500).json({ error });
      }
    }
  }
}

export default new ImageController();
